# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching Fluent in container mode.

Examples
--------

>>> from ansys.fluent.core.launcher.launcher import create_launcher
>>> from ansys.fluent.core.launcher.launch_options import LaunchMode, FluentMode

>>> container_meshing_launcher = create_launcher(LaunchMode.CONTAINER, mode=FluentMode.MESHING)
>>> container_meshing_session = container_meshing_launcher()

>>> container_solver_launcher = create_launcher(LaunchMode.CONTAINER)
>>> container_solver_session = container_solver_launcher()
"""

import inspect
import logging
import os
import time
from typing import Any

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.launcher.fluent_container import (
    configure_container_dict,
    start_fluent_container,
)
from ansys.fluent.core.launcher.launch_options import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
    _get_argvals_and_session,
)
from ansys.fluent.core.launcher.launcher_utils import is_compose
from ansys.fluent.core.launcher.process_launch_string import (
    _build_fluent_launch_args_string,
)
import ansys.fluent.core.launcher.watchdog as watchdog
from ansys.fluent.core.session import _parse_server_info_file
from ansys.fluent.core.utils.fluent_version import FluentVersion

_THIS_DIR = os.path.dirname(__file__)
_OPTIONS_FILE = os.path.join(_THIS_DIR, "fluent_launcher_options.json")
logger = logging.getLogger("pyfluent.launcher")


def _get_server_info_from_container(config_dict):
    """Retrieve the server info from a specified file in a container."""

    host_server_info_file = config_dict["host_server_info_file"]

    time_limit = 0
    while not host_server_info_file.exists():
        time.sleep(2)
        time_limit += 2
        if time_limit > 60:
            raise FileNotFoundError(f"{host_server_info_file} not found.")

    return _parse_server_info_file(str(host_server_info_file))


class DockerLauncher:
    """Instantiates Fluent session in container mode."""

    def __init__(
        self,
        mode: FluentMode | str | None = None,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str | None
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        container_dict: dict | None = None,
        dry_run: bool = False,
        cleanup_on_exit: bool = True,
        start_transcript: bool = True,
        py: bool | None = None,
        gpu: bool | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
        certificates_folder: str | None = None,
        insecure_mode: bool = False,
    ):
        """
        Launch a Fluent session in container mode.

        Parameters
        ----------
        mode : FluentMode
            Specifies the launch mode of Fluent to target a specific session type.
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version :  FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            any of ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values include ``Dimension.TWO``,
            ``Dimension.THREE``, or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values include ``Precision.SINGLE``,
            ``Precision.DOUBLE``, or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        start_timeout : int, optional
            Maximum allowable time in seconds for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        container_dict : dict, optional
            Configuration dictionary for launching Fluent inside a Docker container. See also
            :mod:`~ansys.fluent.core.launcher.fluent_container`.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. If dry running a
            container start, this method will return the configured ``container_dict``. Defaults to False.
        cleanup_on_exit : bool
            Determines whether to shut down the connected Fluent session upon exit or when calling
            the session's `exit()` method. Defaults to True.
        start_transcript : bool
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        start_watchdog : bool, optional
            If True and `cleanup_on_exit` is True, an independent watchdog process is run to ensure that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any, optional
            Service for uploading/downloading files to/from the server.
        certificates_folder : str, optional
            Path to the folder containing TLS certificates for Fluent's gRPC server.
        insecure_mode : bool, optional
            If True, Fluent's gRPC server will be started in insecure mode without TLS.
            This mode is not recommended. For more details on the implications
            and usage of insecure mode, refer to the Fluent documentation.

        Returns
        -------
        Meshing | PureMeshing | Solver | SolverIcing | dict
            Session object or configuration dictionary if ``dry_run`` is True.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        if certificates_folder is None and not insecure_mode:
            raise ValueError(
                "To launch Fluent in secure gRPC mode, set `certificates_folder`."
            )
        if certificates_folder is not None and insecure_mode:
            raise ValueError(
                "`certificates_folder` and `insecure_mode` cannot be set at the same time."
            )

        locals_ = locals().copy()
        argvals = {
            arg: locals_.get(arg)
            for arg in inspect.getargvalues(inspect.currentframe()).args
        }
        self.argvals, self.new_session = _get_argvals_and_session(argvals)
        if self.argvals["start_timeout"] is None:
            self.argvals["start_timeout"] = 60
        self.file_transfer_service = file_transfer_service
        if self.argvals["mode"] == FluentMode.SOLVER_ICING:
            self.argvals["fluent_icing"] = True
        if self.argvals["container_dict"] is None:
            self.argvals["container_dict"] = {}
        if self.argvals["product_version"]:
            self.argvals["container_dict"][
                "image_tag"
            ] = f"v{FluentVersion(self.argvals['product_version']).value}"

        self._args = _build_fluent_launch_args_string(**self.argvals).split()
        if FluentMode.is_meshing(self.argvals["mode"]):
            self._args.append(" -meshing")
        fluent_image_tag = os.getenv("FLUENT_IMAGE_TAG")
        # There is an issue in passing gRPC arguments to Fluent image version 24.1.0 during github runs.
        if self.argvals["insecure_mode"] and fluent_image_tag != "v24.1.0":
            self._args.append(" -grpc-allow-remote-host")
            self._args.append(" -grpc-insecure-mode")
        elif self.argvals["certificates_folder"]:
            self.argvals["container_dict"]["certificates_folder"] = self.argvals[
                "certificates_folder"
            ]
            self._args.append(" -grpc-allow-remote-host")
            self._args.append(" -grpc-certs-folder=/tmp/certs")

    def __call__(self):
        if self.argvals["dry_run"]:
            config_dict, *_ = configure_container_dict(
                self._args, **self.argvals["container_dict"]
            )
            from pprint import pprint

            print("\nDocker container run configuration:\n")
            print("config_dict = ")
            if os.getenv("PYFLUENT_HIDE_LOG_SECRETS") != "1":
                pprint(config_dict)
            else:
                config_dict_h = config_dict.copy()
                config_dict_h.pop("environment")
                pprint(config_dict_h)
                del config_dict_h
            return config_dict

        if is_compose():
            port, config_dict, container = start_fluent_container(
                self._args, self.argvals["container_dict"]
            )

            _, _, password = _get_server_info_from_container(config_dict=config_dict)
        else:
            port, password, container = start_fluent_container(
                self._args, self.argvals["container_dict"]
            )

        fluent_connection = FluentConnection(
            port=port,
            password=password,
            allow_remote_host=self.argvals["insecure_mode"]
            or self.argvals["certificates_folder"] is not None,
            certificates_folder=self.argvals["certificates_folder"],
            insecure_mode=self.argvals["insecure_mode"],
            file_transfer_service=self.file_transfer_service,
            cleanup_on_exit=self.argvals["cleanup_on_exit"],
            slurm_job_id=self.argvals and self.argvals.get("slurm_job_id"),
            inside_container=True,
            container=container,
        )

        session = self.new_session(
            fluent_connection=fluent_connection,
            scheme_eval=fluent_connection._connection_interface.scheme_eval,
            file_transfer_service=self.file_transfer_service,
            start_transcript=self.argvals["start_transcript"],
        )

        session._container = container

        if not is_compose():
            if (
                self.argvals["start_watchdog"] is None
                and self.argvals["cleanup_on_exit"]
            ):
                self.argvals["start_watchdog"] = True
            if self.argvals["start_watchdog"]:
                logger.debug("Launching Watchdog for Fluent container...")
                watchdog.launch(os.getpid(), port, password)

        return session



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_methods(mixing_elbow_settings_session):
    solver = mixing_elbow_settings_session
    solver.setup.models.multiphase.models = "vof"
    solver.setup.general.operating_conditions.gravity = {
        "enable": True,
        "components": [0.0, 0.0, -9.81],
    }
    solver.setup.general.solver.time = "steady"

    p_v_coupling = solver.solution.methods.p_v_coupling
    p_v_coupling.flow_scheme = "Coupled"
    p_v_coupling.coupled_form = False
    assert p_v_coupling() == {
        "flow_scheme": "Coupled",
        "coupled_form": False,
    }
    solver.solution.methods.discretization_scheme = {"pressure": "presto!"}
    assert solver.solution.methods.discretization_scheme() == {
        "mom": "second-order-upwind",
        "omega": "second-order-upwind",
        "mp": "compressive",
        "pressure": "presto!",
        "k": "second-order-upwind",
        "temperature": "second-order-upwind",
    }
    solver.solution.methods.gradient_scheme = "least-square-cell-based"
    assert solver.solution.methods.gradient_scheme() == "least-square-cell-based"

    enable_warped_face = solver.solution.methods.warped_face_gradient_correction
    enable_warped_face(enable=True, mode="fast")
    enable_warped_face(enable=False, mode="fast")

    solver.solution.methods.expert.numerics_pbns.velocity_formulation = "relative"
    assert (
        solver.solution.methods.expert.numerics_pbns.velocity_formulation()
        == "relative"
    )
    solver.solution.methods.expert.numerics_pbns = {
        "implicit_bodyforce_treatment": True,
        "velocity_formulation": "absolute",
        "physical_velocity_formulation": True,
        "disable_rhie_chow_flux": True,
        "presto_pressure_scheme": False,
        "first_to_second_order_blending": 1.0,
    }
    assert solver.solution.methods.expert.numerics_pbns() == {
        "implicit_bodyforce_treatment": True,
        "velocity_formulation": "absolute",
        "physical_velocity_formulation": True,
        "disable_rhie_chow_flux": True,
        "presto_pressure_scheme": False,
        "first_to_second_order_blending": 1.0,
    }
    solver.solution.methods.expert.numerics_pbns.presto_pressure_scheme = True
    assert solver.solution.methods.expert.numerics_pbns.presto_pressure_scheme() is True
    solver.solution.methods.gradient_scheme = "green-gauss-node-based"
    assert solver.solution.methods.gradient_scheme() == "green-gauss-node-based"
    solver.solution.methods.warped_face_gradient_correction(
        enable=True, mode="memory-saving"
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for context managers used in PyFluent."""

from contextlib import contextmanager
from threading import local

_thread_local = local()


def _get_stack():
    if not hasattr(_thread_local, "stack"):
        _thread_local.stack = []
    return _thread_local.stack


@contextmanager
def using(session):
    """Context manager to use a Fluent session."""
    stack = _get_stack()
    stack.append(session)
    try:
        yield
    finally:
        stack.pop()


def _get_active_session():
    stack = _get_stack()
    if stack:
        return stack[-1]


class ReadCase:
    """Context manager to read a Fluent case file."""

    def __init__(self, session=None, case_file: str | None = None):
        self.session = session or _get_active_session()
        self.case_file = case_file

    def __enter__(self):
        if not self.case_file:
            raise ValueError("A case file path must be provided to read.")
        self.session.settings.file.read_case(file_name=self.case_file)
        return self.session

    def __exit__(self, exc_type, exc_value, traceback):
        pass



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any

import pytest

from ansys.fluent.core.examples import download_file
from ansys.fluent.core.services.reduction import _locn_names_and_objs
from ansys.fluent.core.solver.function import reduction


def _test_locn_extraction(solver1, solver2):
    solver1_boundary_conditions = solver1.setup.boundary_conditions
    solver2_boundary_conditions = solver2.setup.boundary_conditions
    locns = _locn_names_and_objs(["inlet1"])
    assert locns == [["inlet1", ["inlet1"]]]

    all_bcs = solver1_boundary_conditions
    locns = _locn_names_and_objs(all_bcs)
    assert locns == [
        ["interior--fluid", all_bcs],
        ["outlet", all_bcs],
        ["inlet1", all_bcs],
        ["inlet2", all_bcs],
        ["wall", all_bcs],
    ]

    locns = _locn_names_and_objs([all_bcs["inlet1"]])
    assert locns == [["inlet1", all_bcs["inlet1"]]]

    all_bcs = solver1_boundary_conditions
    all_bcs2 = solver2_boundary_conditions
    locns = _locn_names_and_objs([all_bcs, all_bcs2])
    assert locns == [
        ["interior--fluid", all_bcs],
        ["outlet", all_bcs],
        ["inlet1", all_bcs],
        ["inlet2", all_bcs],
        ["wall", all_bcs],
        ["interior--fluid", all_bcs2],
        ["outlet", all_bcs2],
        ["inlet1", all_bcs2],
        ["inlet2", all_bcs2],
        ["wall", all_bcs2],
    ]


def _test_context(solver):
    solver.solution.initialization.hybrid_initialize()

    assert solver.fields.reduction.area(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]],
        ctxt=solver,
    )

    assert solver.fields.reduction.area(locations=["inlet1"], ctxt=solver)


def _test_area_average(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = (
        "AreaAve(AbsolutePressure, ['inlet1'])"
    )
    expr_val = solver_named_expressions["test_expr_1"].get_value()
    assert isinstance(expr_val, float) and expr_val != 0.0
    val = solver.fields.reduction.area_average(
        expression="AbsolutePressure",
        locations=solver.setup.boundary_conditions.velocity_inlet,
    )
    assert val == expr_val
    solver_named_expressions.pop(key="test_expr_1")


def _test_min(solver1, solver2):
    s1_min = solver1.fields.reduction.minimum(
        expression="AbsolutePressure",
        locations=[solver1.setup.boundary_conditions.velocity_inlet],
    )
    s2_min = solver2.fields.reduction.minimum(
        expression="AbsolutePressure",
        locations=[solver2.setup.boundary_conditions.velocity_inlet],
    )
    result = reduction.minimum(
        expression="AbsolutePressure",
        locations=[
            solver1.setup.boundary_conditions.velocity_inlet,
            solver2.setup.boundary_conditions.velocity_inlet,
        ],
    )
    assert result == min(s1_min, s2_min)


def _test_count(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    s_velocity_inlet = solver.setup.boundary_conditions.velocity_inlet
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = "Count(['inlet1'])"
    expr_val_1 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = "Count(['inlet2'])"
    expr_val_2 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = "Count(['inlet1', 'inlet2'])"
    expr_val_3 = solver_named_expressions["test_expr_1"].get_value()
    assert expr_val_3 == expr_val_1 + expr_val_2
    red_val_1 = solver.fields.reduction.count(locations=[s_velocity_inlet["inlet1"]])
    red_val_2 = solver.fields.reduction.count(locations=[s_velocity_inlet["inlet2"]])
    red_val_3 = solver.fields.reduction.count(locations=[s_velocity_inlet])
    assert red_val_1 == expr_val_1
    assert red_val_2 == expr_val_2
    assert red_val_3 == expr_val_3
    solver_named_expressions.pop(key="test_expr_1")


def _test_count_if(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    s_velocity_inlet = solver.setup.boundary_conditions.velocity_inlet
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = (
        "CountIf(AbsolutePressure > 0[Pa], ['inlet1'])"
    )
    expr_val_1 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = (
        "CountIf(AbsolutePressure > 0[Pa], ['inlet2'])"
    )
    expr_val_2 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = (
        "CountIf(AbsolutePressure > 0[Pa], ['inlet1', 'inlet2'])"
    )
    expr_val_3 = solver_named_expressions["test_expr_1"].get_value()
    assert expr_val_3 == expr_val_1 + expr_val_2
    red_val_1 = solver.fields.reduction.count_if(
        condition="AbsolutePressure > 0[Pa]", locations=["inlet1"]
    )
    red_val_2 = solver.fields.reduction.count_if(
        condition="AbsolutePressure > 0[Pa]", locations=[s_velocity_inlet["inlet2"]]
    )
    red_val_3 = solver.fields.reduction.count_if(
        condition="AbsolutePressure > 0[Pa]", locations=[s_velocity_inlet]
    )
    assert red_val_1 == expr_val_1
    assert red_val_2 == expr_val_2
    assert red_val_3 == expr_val_3
    solver_named_expressions.pop(key="test_expr_1")


def _test_centroid(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    velocity_inlet = solver.setup.boundary_conditions.velocity_inlet
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = "Centroid(['inlet1'])"
    expr_val_1 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = "Centroid(['inlet2'])"
    expr_val_2 = solver_named_expressions["test_expr_1"].get_value()
    solver_named_expressions["test_expr_1"].definition = (
        "Centroid(['inlet1', 'inlet2'])"
    )
    expr_val_3 = solver_named_expressions["test_expr_1"].get_value()
    red_val_1 = solver.fields.reduction.centroid(locations=[velocity_inlet["inlet1"]])
    red_val_2 = solver.fields.reduction.centroid(locations=[velocity_inlet["inlet2"]])
    red_val_3 = solver.fields.reduction.centroid(locations=[velocity_inlet])
    assert [red_val_1[0], red_val_1[1], red_val_1[2]] == expr_val_1
    assert [red_val_2[0], red_val_2[1], red_val_2[2]] == expr_val_2
    assert [red_val_3[0], red_val_3[1], red_val_3[2]] == expr_val_3
    solver_named_expressions.pop(key="test_expr_1")


def _test_area_integrated_average(solver1, solver2):
    solver1.solution.initialization.hybrid_initialize()
    solver2.solution.initialization.hybrid_initialize()
    solver1_boundary_conditions = solver1.setup.boundary_conditions
    solver2_boundary_conditions = solver2.setup.boundary_conditions
    solver1_named_expr = solver1.setup.named_expressions
    solver2_named_expr = solver2.setup.named_expressions

    solver1_named_expr["test_expr_1"] = {}
    solver1_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet1'])"
    )
    expr_val_1 = solver1_named_expr["test_expr_1"].get_value()

    solver1_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet2'])"
    )
    expr_val_2 = solver1_named_expr["test_expr_1"].get_value()
    solver1_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet1', 'inlet2'])"
    )
    expr_val_3 = solver1_named_expr["test_expr_1"].get_value()

    assert expr_val_3 - (expr_val_1 + expr_val_2) <= 0.000000001

    red_val_1 = solver1.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1_boundary_conditions.velocity_inlet["inlet1"]],
    )
    red_val_2 = solver1.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1_boundary_conditions.velocity_inlet["inlet2"]],
    )
    red_val_3 = solver1.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1_boundary_conditions.velocity_inlet],
    )

    assert red_val_1 == expr_val_1
    assert red_val_2 == expr_val_2
    assert red_val_3 == expr_val_3

    solver2_named_expr["test_expr_1"] = {}
    solver2_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet1'])"
    )
    expr_val_4 = solver2_named_expr["test_expr_1"].get_value()

    solver2_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet2'])"
    )
    expr_val_5 = solver2_named_expr["test_expr_1"].get_value()
    solver2_named_expr["test_expr_1"].definition = (
        "AreaInt(AbsolutePressure, ['inlet1', 'inlet2'])"
    )
    expr_val_6 = solver2_named_expr["test_expr_1"].get_value()

    assert expr_val_6 - (expr_val_4 + expr_val_5) <= 0.000000001

    red_val_4 = solver2.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2_boundary_conditions.velocity_inlet["inlet1"]],
    )
    red_val_5 = solver2.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2_boundary_conditions.velocity_inlet["inlet2"]],
    )
    red_val_6 = solver2.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2_boundary_conditions.velocity_inlet],
    )

    assert red_val_4 == expr_val_4
    assert red_val_5 == expr_val_5
    assert red_val_6 == expr_val_6

    red_val_7 = solver2.fields.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[
            solver1_boundary_conditions.velocity_inlet,
            solver2_boundary_conditions.velocity_inlet,
        ],
    )

    assert red_val_7 - (expr_val_3 + expr_val_6) <= 0.000000001

    solver1_named_expr.pop(key="test_expr_1")


def _test_error_handling(solver):
    if int(solver._version) < 241:
        with pytest.raises(RuntimeError):
            solver.fields.reduction.area_average(
                expression="AbsoluteVelocity",  # This is a wrong expression intentionally passed
                locations=solver.setup.boundary_conditions.velocity_inlet,
            )


def _test_force(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = "Force(['wall'])"
    expr_val_1 = solver_named_expressions["test_expr_1"].get_value()

    red_total_force = solver.fields.reduction.force(
        locations=[solver.setup.boundary_conditions.wall]
    )
    red_pressure_force = solver.fields.reduction.pressure_force(
        locations=[solver.setup.boundary_conditions.wall]
    )
    red_viscous_force = solver.fields.reduction.viscous_force(
        locations=[solver.setup.boundary_conditions.wall]
    )

    assert [red_total_force[0], red_total_force[1], red_total_force[2]] == expr_val_1

    assert red_pressure_force[0] + red_viscous_force[0] == red_total_force[0]

    assert red_pressure_force[1] + red_viscous_force[1] == red_total_force[1]

    assert red_pressure_force[2] + red_viscous_force[2] == red_total_force[2]

    solver_named_expressions.pop(key="test_expr_1")


def _test_moment(solver):
    solver.solution.initialization.hybrid_initialize()
    solver_named_expressions = solver.setup.named_expressions
    location = solver.setup.boundary_conditions.wall
    solver_named_expressions["test_expr_1"] = {}
    solver_named_expressions["test_expr_1"].definition = (
        "Moment(Force(['wall']),['wall'])"
    )
    expr_val_1 = solver_named_expressions["test_expr_1"].get_value()

    solver_named_expressions["test_expr_1"].definition = "Moment(['inlet1'],['wall'])"
    expr_val_2 = solver_named_expressions["test_expr_1"].get_value()

    red_moment_force = solver.fields.reduction.moment(
        expression="Force(['wall'])", locations=[location]
    )

    red_moment_location = solver.fields.reduction.moment(
        expression="['inlet1']", locations=[location]
    )

    assert [red_moment_force[0], red_moment_force[1], red_moment_force[2]] == expr_val_1
    assert [
        red_moment_location[0],
        red_moment_location[1],
        red_moment_location[2],
    ] == expr_val_2

    solver_named_expressions.pop(key="test_expr_1")


def _test_sum(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions["test_expr_1"].definition = (
        "Sum(AbsolutePressure, ['inlet1'], Weight=Area)"
    )
    expr_val = solver.setup.named_expressions["test_expr_1"].get_value()
    assert isinstance(expr_val, float) and expr_val != 0.0

    val = solver.fields.reduction.sum(
        expression="AbsolutePressure",
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]],
        weight="Area",
    )

    assert val == expr_val
    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_sum_if(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions["test_expr_1"].definition = (
        "SumIf(AbsolutePressure, AbsolutePressure > 0[Pa], ['inlet1'], Weight=Area)"
    )
    expr_val = solver.setup.named_expressions["test_expr_1"].get_value()
    assert isinstance(expr_val, float) and expr_val != 0.0

    val = solver.fields.reduction.sum_if(
        expression="AbsolutePressure",
        condition="AbsolutePressure > 0[Pa]",
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]],
        weight="Area",
    )

    assert val == expr_val
    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_centroid_2_sources(solver1, solver2):
    s1_cent = solver1.fields.reduction.centroid(
        locations=[solver1.setup.boundary_conditions.velocity_inlet]
    )
    s2_cent = solver2.fields.reduction.centroid(
        locations=[solver2.setup.boundary_conditions.velocity_inlet]
    )

    result = reduction.centroid(
        locations=[
            solver1.setup.boundary_conditions.velocity_inlet,
            solver2.setup.boundary_conditions.velocity_inlet,
        ]
    )
    assert [round(x, 5) for x in result] == [
        (round(x, 5) + round(y, 5)) / 2 for x, y in zip(*[s1_cent, s2_cent])
    ]


@pytest.fixture
def static_mixer_case_session2(static_mixer_case_session: Any):
    return static_mixer_case_session


@pytest.mark.fluent_version(">=24.1")
def test_reductions(
    static_mixer_case_session: Any, static_mixer_case_session2: Any
) -> None:
    solver1 = static_mixer_case_session
    solver2 = static_mixer_case_session2
    _test_context(solver1)
    _test_locn_extraction(solver1, solver2)
    _test_area_average(solver1)
    _test_count(solver1)
    _test_count_if(solver1)
    _test_centroid(solver1)
    _test_area_integrated_average(solver1, solver2)
    _test_error_handling(solver1)
    _test_force(solver1)
    _test_moment(solver1)
    _test_sum(solver1)
    _test_sum_if(solver1)
    # The case and data are changed after this point to check the functional reduction with multiple solvers
    case_path = download_file(
        file_name="exhaust_system.cas.h5", directory="pyfluent/exhaust_system"
    )
    download_file(
        file_name="exhaust_system.dat.h5", directory="pyfluent/exhaust_system"
    )
    solver1.file.read_case_data(file_name=case_path)
    case_path1 = download_file("elbow1.cas.h5", "pyfluent/file_session")
    download_file("elbow1.dat.h5", "pyfluent/file_session")
    solver2.file.read_case_data(file_name=case_path1)
    _test_min(solver1, solver2)
    _test_centroid_2_sources(solver1, solver2)


@pytest.mark.fluent_version(">=24.2")
def test_reduction_does_not_modify_case(static_mixer_case_session: Any):
    solver = static_mixer_case_session
    solver.solution.initialization.hybrid_initialize()
    # After reading the static-mixer case in Fluent, case-modifed? flag is somehow True
    solver.scheme.eval("(%save-case-id)")
    assert not solver.scheme.eval("(case-modified?)")
    solver.reduction.area_average(
        expression="AbsolutePressure",
        locations=solver.setup.boundary_conditions.velocity_inlet,
    )
    assert not solver.scheme.eval("(case-modified?)")


@pytest.mark.fluent_version(">=24.2")
def test_fix_for_invalid_location_inputs(static_mixer_case_session: Any):
    solver = static_mixer_case_session
    solver.solution.initialization.hybrid_initialize()

    assert solver.fields.reduction.area(locations=["inlet1"], ctxt=solver)

    with pytest.raises(ValueError):
        assert solver.fields.reduction.area(locations=["inlet-1"], ctxt=solver)

    with pytest.raises(KeyError):
        assert solver.fields.reduction.area(
            locations=[solver.setup.boundary_conditions.velocity_inlet["inlet-1"]]
        )

    assert solver.fields.reduction.area(locations=["inlet1"])

    with pytest.raises(ValueError):
        assert solver.fields.reduction.area(locations=["inlet-1"])


@pytest.mark.fluent_version(">=25.2")
def test_fix_for_empty_location_inputs(static_mixer_case_session: Any):
    solver = static_mixer_case_session
    solver.solution.initialization.hybrid_initialize()

    assert solver.fields.reduction.area(locations=["inlet1"])

    with pytest.raises(RuntimeError):
        assert reduction.area(locations=[], ctxt=solver)

    with pytest.raises(RuntimeError):
        assert reduction.area_average(
            expression="AbsolutePressure", locations=[], ctxt=solver
        )

    with pytest.raises(RuntimeError):
        assert reduction.centroid(locations=[], ctxt=solver)

    with pytest.raises(RuntimeError):
        assert solver.fields.reduction.area(locations=[])

    with pytest.raises(RuntimeError):
        assert solver.fields.reduction.area_average(
            expression="AbsolutePressure", locations=[]
        )

    with pytest.raises(RuntimeError):
        assert solver.fields.reduction.centroid(locations=[])



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching Fluent in standalone mode.

Examples
--------

>>> from ansys.fluent.core.launcher.launcher import create_launcher
>>> from ansys.fluent.core.launcher.launch_options import LaunchMode, FluentMode

>>> standalone_meshing_launcher = create_launcher(LaunchMode.STANDALONE, mode=FluentMode.MESHING)
>>> standalone_meshing_session = standalone_meshing_launcher()

>>> standalone_solver_launcher = create_launcher(LaunchMode.STANDALONE)
>>> standalone_solver_session = standalone_solver_launcher()
"""

import inspect
import logging
import os
from pathlib import Path
import subprocess
from typing import Any, Dict

from ansys.fluent.core.launcher.error_handler import (
    LaunchFluentError,
    _raise_non_gui_exception_in_windows,
)
from ansys.fluent.core.launcher.launch_options import (
    Dimension,
    FluentLinuxGraphicsDriver,
    FluentMode,
    FluentWindowsGraphicsDriver,
    Precision,
    UIMode,
    _get_argvals_and_session,
    _get_standalone_launch_fluent_version,
)
from ansys.fluent.core.launcher.launcher_utils import (
    _await_fluent_launch,
    _build_journal_argument,
    _confirm_watchdog_start,
    _get_subprocess_kwargs_for_fluent,
    is_windows,
)
from ansys.fluent.core.launcher.process_launch_string import _generate_launch_string
from ansys.fluent.core.launcher.server_info import (
    _get_server_info,
    _get_server_info_file_names,
)
import ansys.fluent.core.launcher.watchdog as watchdog
from ansys.fluent.core.utils.fluent_version import FluentVersion

logger = logging.getLogger("pyfluent.launcher")


class StandaloneLauncher:
    """Instantiates Fluent session in standalone mode."""

    def __init__(
        self,
        mode: FluentMode | str | None = None,
        ui_mode: UIMode | str | None = None,
        graphics_driver: (
            FluentWindowsGraphicsDriver | FluentLinuxGraphicsDriver | str
        ) = None,
        product_version: FluentVersion | str | float | int | None = None,
        dimension: Dimension | int | None = None,
        precision: Precision | str | None = None,
        processor_count: int | None = None,
        journal_file_names: None | str | list[str] = None,
        start_timeout: int = 60,
        additional_arguments: str = "",
        env: Dict[str, Any] | None = None,
        cleanup_on_exit: bool = True,
        dry_run: bool = False,
        start_transcript: bool = True,
        case_file_name: str | None = None,
        case_data_file_name: str | None = None,
        lightweight_mode: bool | None = None,
        py: bool | None = None,
        gpu: bool | None = None,
        cwd: str | None = None,
        fluent_path: str | None = None,
        topy: str | list | None = None,
        start_watchdog: bool | None = None,
        file_transfer_service: Any | None = None,
    ):
        """
        Launch a Fluent session in standalone mode.

        Parameters
        ----------
        mode : FluentMode
            Specifies the launch mode of Fluent to target a specific session type.
        ui_mode : UIMode
            Defines the user interface mode for Fluent. Options correspond to values in the ``UIMode`` enum.
        graphics_driver : FluentWindowsGraphicsDriver or FluentLinuxGraphicsDriver
            Specifies the graphics driver for Fluent. Options are from the ``FluentWindowsGraphicsDriver`` enum
            (for Windows) or the ``FluentLinuxGraphicsDriver`` enum (for Linux).
        product_version : FluentVersion or str or float or int, optional
            Indicates the version of Ansys Fluent to launch. For example, to use version 2025 R1, pass
            ``FluentVersion.v251``, ``"25.1.0"``, ``"25.1"``, ``25.1``, or ``251``. Defaults to ``None``,
            which uses the newest installed version.
        dimension : Dimension or int, optional
            Specifies the geometric dimensionality of the Fluent simulation. Defaults to ``None``,
            which corresponds to ``Dimension.THREE``. Acceptable values are from the ``Dimension`` enum
            (``Dimension.TWO`` or ``Dimension.THREE``) or integers ``2`` and ``3``.
        precision : Precision or str, optional
            Defines the floating point precision. Defaults to ``None``, which corresponds to
            ``Precision.DOUBLE``. Acceptable values are from the ``Precision`` enum (``Precision.SINGLE``
            or ``Precision.DOUBLE``) or strings ``"single"`` and ``"double"``.
        processor_count : int, optional
            Specifies the number of processors to use. Defaults to ``None``, which uses 1 processor.
            In job scheduler environments, this value limits the total number of allocated cores.
        journal_file_names : str or list of str, optional
            Path(s) to a Fluent journal file(s) that Fluent will execute. Defaults to ``None``.
        start_timeout : int, optional
            Maximum time in seconds allowed for connecting to the Fluent server. Defaults to 60 seconds.
        additional_arguments : str, optional
            Additional command-line arguments for Fluent, formatted as they would be on the command line.
        env : dict[str, str], optional
            A mapping for modifying environment variables in Fluent. Defaults to ``None``.
        cleanup_on_exit : bool, optional
            Determines whether to shut down the connected Fluent session when exiting PyFluent or calling
            the session's `exit()` method. Defaults to True.
        dry_run : bool, optional
            If True, does not launch Fluent but prints configuration information instead. The `call()` method
            returns a tuple containing the launch string and server info file name. Defaults to False.
        start_transcript : bool, optional
            Indicates whether to start streaming the Fluent transcript in the client. Defaults to True;
            streaming can be controlled via `transcript.start()` and `transcript.stop()` methods on the session object.
        case_file_name : str, optional
            Name of the case file to read into the Fluent session. Defaults to None.
        case_data_file_name : str, optional
            Name of the case data file. If both case and data files are provided, they are read into the session.
        lightweight_mode : bool, optional
            If True, runs in lightweight mode where mesh settings are read into a background solver session,
            replacing it once complete. This parameter is only applicable when `case_file_name` is provided; defaults to False.
        py : bool, optional
            If True, runs Fluent in Python mode. Defaults to None.
        gpu : bool, optional
            If True, starts Fluent with GPU Solver enabled.
        cwd : str, optional
            Working directory for the Fluent client.
        fluent_path: str, optional
            User-specified path for Fluent installation.
        topy :  bool or str, optional
            A flag indicating whether to write equivalent Python journals from provided journal files; can also specify
            a filename for the new Python journal.
        start_watchdog : bool, optional
            When `cleanup_on_exit` is True, defaults to True; an independent watchdog process ensures that any local
            GUI-less Fluent sessions started by PyFluent are properly closed when the current Python process ends.
        file_transfer_service : Any
            Service for uploading/downloading files to/from the server.

        Raises
        ------
        UnexpectedKeywordArgument
            If an unexpected keyword argument is provided.

        Notes
        -----
        In job scheduler environments (e.g., SLURM, LSF, PBS), resources and compute nodes are allocated,
        and core counts are queried from these environments before being passed to Fluent.
        """
        import ansys.fluent.core as pyfluent

        locals_ = locals().copy()
        argvals = {
            arg: locals_.get(arg)
            for arg in inspect.getargvalues(inspect.currentframe()).args
        }
        self.argvals, self.new_session = _get_argvals_and_session(argvals)
        self.file_transfer_service = file_transfer_service
        if os.getenv("PYFLUENT_SHOW_SERVER_GUI") == "1":
            ui_mode = UIMode.GUI
        self.argvals["ui_mode"] = UIMode(ui_mode)
        if self.argvals["start_timeout"] is None:
            self.argvals["start_timeout"] = 60
        if self.argvals["lightweight_mode"] is None:
            self.argvals["lightweight_mode"] = False
        fluent_version = _get_standalone_launch_fluent_version(self.argvals)
        if fluent_version:
            _raise_non_gui_exception_in_windows(self.argvals["ui_mode"], fluent_version)

        if (
            fluent_version
            and fluent_version >= FluentVersion.v251
            and self.argvals["py"] is None
        ):
            self.argvals["py"] = True

        if os.getenv("PYFLUENT_FLUENT_DEBUG") == "1":
            self.argvals["fluent_debug"] = True

        server_info_file_name_for_server, server_info_file_name_for_client = (
            _get_server_info_file_names()
        )
        self._server_info_file_name = server_info_file_name_for_client
        self._launch_string = _generate_launch_string(
            self.argvals,
            server_info_file_name_for_server,
        )

        self._sifile_last_mtime = Path(self._server_info_file_name).stat().st_mtime
        self._kwargs = _get_subprocess_kwargs_for_fluent(
            self.argvals["env"], self.argvals
        )
        if self.argvals["cwd"]:
            self._kwargs.update(cwd=self.argvals["cwd"])
        self._launch_string += _build_journal_argument(
            self.argvals["topy"], self.argvals["journal_file_names"]
        )

        if is_windows():
            if pyfluent.LAUNCH_FLUENT_STDOUT or pyfluent.LAUNCH_FLUENT_STDERR:
                self._launch_cmd = self._launch_string
            else:
                # Using 'start.exe' is better; otherwise Fluent is more susceptible to bad termination attempts.
                self._launch_cmd = 'start "" ' + self._launch_string
        else:
            if self.argvals["ui_mode"] not in [UIMode.GUI, UIMode.HIDDEN_GUI]:
                # Using nohup to hide Fluent output from the current terminal
                self._launch_cmd = "nohup " + self._launch_string + " &"
            else:
                self._launch_cmd = self._launch_string

    def __call__(self):
        if self.argvals["dry_run"]:
            print(f"Fluent launch string: {self._launch_string}")
            return self._launch_string, self._server_info_file_name
        try:
            logger.debug(f"Launching Fluent with command: {self._launch_cmd}")

            process = subprocess.Popen(self._launch_cmd, **self._kwargs)

            try:
                _await_fluent_launch(
                    self._server_info_file_name,
                    self.argvals["start_timeout"],
                    self._sifile_last_mtime,
                )
            except TimeoutError as ex:
                if is_windows():
                    logger.warning(f"Exception caught - {type(ex).__name__}: {ex}")
                    launch_cmd = self._launch_string.replace('"', "", 2)
                    self._kwargs.update(shell=False)
                    logger.warning(
                        f"Retrying Fluent launch with less robust command: {launch_cmd}"
                    )
                    process = subprocess.Popen(launch_cmd, **self._kwargs)
                    _await_fluent_launch(
                        self._server_info_file_name,
                        self.argvals["start_timeout"],
                        self._sifile_last_mtime,
                    )
                else:
                    raise ex

            session = self.new_session._create_from_server_info_file(
                server_info_file_name=self._server_info_file_name,
                file_transfer_service=self.file_transfer_service,
                cleanup_on_exit=self.argvals["cleanup_on_exit"],
                start_transcript=self.argvals["start_transcript"],
                launcher_args=self.argvals,
                inside_container=False,
            )
            session._process = process
            start_watchdog = _confirm_watchdog_start(
                self.argvals["start_watchdog"],
                self.argvals["cleanup_on_exit"],
                session._fluent_connection,
            )
            if start_watchdog:
                logger.info("Launching Watchdog for local Fluent client...")
                values = _get_server_info(self._server_info_file_name)
                if len(values) == 3:
                    ip, port, password = values
                    watchdog.launch(os.getpid(), port, password, ip)
            if self.argvals["case_file_name"]:
                if FluentMode.is_meshing(self.argvals["mode"]):
                    session.tui.file.read_case(self.argvals["case_file_name"])
                elif self.argvals["lightweight_mode"]:
                    session.read_case_lightweight(self.argvals["case_file_name"])
                else:
                    session.file.read(
                        file_type="case",
                        file_name=self.argvals["case_file_name"],
                    )
            if self.argvals["case_data_file_name"]:
                if not FluentMode.is_meshing(self.argvals["mode"]):
                    session.file.read(
                        file_type="case-data",
                        file_name=self.argvals["case_data_file_name"],
                    )
                else:
                    raise RuntimeError(
                        "Case and data file cannot be read in meshing mode."
                    )

            return session
        except Exception as ex:
            logger.error(f"Exception caught - {type(ex).__name__}: {ex}")
            raise LaunchFluentError(self._launch_cmd) from ex
        finally:
            server_info_file = Path(self._server_info_file_name)
            if server_info_file.exists():
                server_info_file.unlink()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.mark.settings_only
@pytest.mark.fluent_version("latest")
def test_controls(mixing_elbow_settings_session):
    solver = mixing_elbow_settings_session
    solver.setup.models.multiphase.models = "vof"
    assert solver.setup.models.multiphase.models() == "vof"
    solver.setup.general.operating_conditions.gravity = {
        "enable": True,
        "components": [0.0, 0.0, -9.81],
    }
    assert solver.setup.general.operating_conditions.gravity.components() == [
        0,
        0,
        -9.81,
    ]
    solver.setup.general.solver.time = "steady"
    assert solver.setup.general.solver.time() == "steady"
    param_coarsening = (
        solver.solution.controls.advanced.multi_grid.amg_controls.coupled_parameters.coarsening_parameters
    )
    param_coarsening.laplace_coarsening = True
    assert param_coarsening() == {
        "max_coarse_levels": 40,
        "coarsen_by_interval": 8,
        "conservative_coarsening": False,
        "aggressive_coarsening": False,
        "laplace_coarsening": True,
    }
    param_coarsening.max_coarse_levels = 45
    assert param_coarsening.max_coarse_levels() == 45

    param_coarsening.set_state(
        {
            "max_coarse_levels": 48,
            "coarsen_by_interval": 9,
            "conservative_coarsening": True,
            "aggressive_coarsening": True,
            "laplace_coarsening": True,
        }
    )
    assert param_coarsening.max_coarse_levels() == 48
    assert param_coarsening.coarsen_by_interval() == 9
    assert param_coarsening.conservative_coarsening() is True
    assert param_coarsening.aggressive_coarsening() is True

    param_fixed_cycle = (
        solver.solution.controls.advanced.multi_grid.amg_controls.scalar_parameters.fixed_cycle_parameters
    )
    param_fixed_cycle.max_cycle = 300
    assert param_fixed_cycle.max_cycle() == 300
    param_fixed_cycle.set_state(
        {
            "pre_sweeps": 1,
            "post_sweeps": 2,
            "max_cycle": 350,
        }
    )
    assert param_fixed_cycle() == {
        "pre_sweeps": 1,
        "post_sweeps": 2,
        "max_cycle": 350,
    }
    solver.solution.methods.p_v_coupling.flow_scheme = "Coupled"
    assert solver.solution.methods.p_v_coupling.flow_scheme() == "Coupled"
    solver.solution.methods.p_v_coupling.coupled_form = True
    assert solver.solution.methods.p_v_coupling.coupled_form() is True
    solver.solution.controls.advanced.multi_grid.amg_controls.scalar_parameters.smoother_type = (
        "Gauss-Seidel"
    )
    assert (
        solver.solution.controls.advanced.multi_grid.amg_controls.scalar_parameters.smoother_type()
        == "Gauss-Seidel"
    )
    assert solver.solution.controls.pseudo_time_explicit_relaxation_factor() == {
        "global_dt_pseudo_relax": {
            "turb-viscosity": 1.0,
            "body-force": 1.0,
            "k": 0.75,
            "omega": 0.75,
            "mp": 0.5,
            "density": 1.0,
            "temperature": 0.75,
        }
    }
    solver.solution.controls.pseudo_time_explicit_relaxation_factor = {
        "global_dt_pseudo_relax": {"turb-viscosity": 0.9, "body-force": 0.8}
    }
    assert solver.solution.controls.pseudo_time_explicit_relaxation_factor() == {
        "global_dt_pseudo_relax": {
            "turb-viscosity": 0.9,
            "body-force": 0.8,
            "k": 0.75,
            "omega": 0.75,
            "mp": 0.5,
            "density": 1.0,
            "temperature": 0.75,
        }
    }
    solver.solution.methods.p_v_coupling.flow_scheme = "SIMPLE"
    solver.solution.controls.under_relaxation = {"pressure": 0.9}
    solver.solution.controls.under_relaxation = {"density": 0.9}
    assert solver.solution.controls.under_relaxation() == {
        "mom": 0.7,
        "turb-viscosity": 1,
        "density": 0.9,
        "omega": 0.8,
        "mp": 0.5,
        "body-force": 1.0,
        "pressure": 0.9,
        "k": 0.8,
        "temperature": 1.0,
    }
    solver.solution.controls.under_relaxation = {
        "body-force": 0.7,
        "density": 0.75,
        "mom": 0.8,
    }
    assert solver.solution.controls.under_relaxation() == {
        "mom": 0.8,
        "turb-viscosity": 1,
        "density": 0.75,
        "omega": 0.8,
        "mp": 0.5,
        "body-force": 0.7,
        "pressure": 0.9,
        "k": 0.8,
        "temperature": 1.0,
    }



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Batch RPC service."""

import inspect
import logging
from types import ModuleType
from typing import TypeVar
import weakref

from google.protobuf.message import Message
import grpc

import ansys.api.fluent.v0 as api
from ansys.api.fluent.v0 import batch_ops_pb2, batch_ops_pb2_grpc

_TBatchOps = TypeVar("_TBatchOps", bound="BatchOps")

network_logger: logging.Logger = logging.getLogger("pyfluent.networking")


class BatchOpsService:
    """Class wrapping methods in batch RPC service."""

    def __init__(self, channel: grpc.Channel, metadata: list[tuple[str, str]]) -> None:
        """__init__ method of BatchOpsService class."""

        from ansys.fluent.core.services.interceptors import GrpcErrorInterceptor

        intercept_channel = grpc.intercept_channel(
            channel,
            GrpcErrorInterceptor(),
        )
        self._stub = batch_ops_pb2_grpc.BatchOpsStub(intercept_channel)
        self._metadata = metadata

    def execute(
        self, request: batch_ops_pb2.ExecuteRequest
    ) -> batch_ops_pb2.ExecuteResponse:
        """Execute RPC of BatchOps service."""
        return self._stub.Execute(request, metadata=self._metadata)


class BatchOps:
    """Class to execute operations in batch in Fluent.

    Examples
    --------
    >>> with pyfluent.BatchOps(solver):
    >>>     solver.tui.file.read_case("mixing_elbow.cas.h5")
    >>>     solver.results.graphics.mesh["mesh-1"] = {}

    Above code will execute both operations through a single gRPC call upon exiting the
    ``with`` block.

    Operations that perform queries in Fluent are executed immediately, while others are
    queued for batch execution. Some queries are executed behind the scenes while
    queueing an operation for batch execution, and we must ensure that they do not
    depend on previously queued operations.


    For example,

    >>> with pyfluent.BatchOps(solver):
    >>>     solver.tui.file.read_case("mixing_elbow.cas.h5")
    >>>     solver.results.graphics.mesh["mesh-1"] = {}
    >>>     solver.results.graphics.mesh["mesh-1"].surfaces_list = ["wall-elbow"]

    will throw a ``KeyError`` as ``solver.results.graphics.mesh["mesh-1"]`` attempts to
    access the ``mesh-1`` mesh object which has not been created yet.
    """

    _proto_files: list[ModuleType] | None = None

    def _instance():
        return None

    @classmethod
    def instance(cls) -> _TBatchOps | None:
        """Get the BatchOps instance.

        Returns
        -------
        BatchOps
            BatchOps instance
        """
        return cls._instance()

    class Op:
        """Class to create a single batch operation."""

        def __init__(
            self, package: str, service: str, method: str, request_body: bytes
        ) -> None:
            """__init__ method of Op class."""
            self._request = batch_ops_pb2.ExecuteRequest(
                package=package,
                service=service,
                method=method,
                request_body=request_body,
            )
            if not BatchOps._proto_files:
                BatchOps._proto_files = [
                    x[1]
                    for x in inspect.getmembers(api, inspect.ismodule)
                    if hasattr(x[1], "DESCRIPTOR")
                ]
            self._supported = False
            self.response_cls = None
            for file in BatchOps._proto_files:
                file_desc = file.DESCRIPTOR
                if file_desc.package == package:
                    service_desc = file_desc.services_by_name.get(service)
                    if service_desc:
                        # TODO Add custom option in .proto files to identify getters
                        if not method.startswith("Get") and not method.startswith(
                            "get"
                        ):
                            method_desc = service_desc.methods_by_name.get(method)
                            if (
                                method_desc
                                and not method_desc.client_streaming
                                and not method_desc.server_streaming
                            ):
                                self._supported = True
                                response_cls_name = method_desc.output_type.name
                                # TODO Get the response_cls from message_factory
                                try:
                                    self.response_cls = getattr(file, response_cls_name)
                                    break
                                except AttributeError:
                                    pass
            if self._supported:
                self._request = batch_ops_pb2.ExecuteRequest(
                    package=package,
                    service=service,
                    method=method,
                    request_body=request_body,
                )
                self._status = None
                self._result = None
            self.queued = False

        def update_result(self, status: batch_ops_pb2.ExecuteStatus, data: str) -> None:
            """Update results after the batch operation is executed."""
            obj = self.response_cls()
            try:
                obj.ParseFromString(data)
            except Exception:
                pass
            self._status = status
            self._result = obj

    def __new__(cls, session) -> _TBatchOps:
        if cls.instance() is None:
            instance = super(BatchOps, cls).__new__(cls)
            instance._service: BatchOpsService = session._batch_ops_service
            instance._ops: list[BatchOps.Op] = []
            instance.batching = False
            cls._instance = weakref.ref(instance)
        return cls.instance()

    def __enter__(self) -> _TBatchOps:
        """Entering the with block."""
        self.clear_ops()
        self.batching = True
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Exiting from the with block."""
        network_logger.debug("Executing batch operations")
        self.batching = False
        if not exc_type:
            requests = (x._request for x in self._ops)
            responses = self._service.execute(requests)
            for i, response in enumerate(responses):
                self._ops[i].update_result(response.status, response.response_body)

    def add_op(self, package: str, service: str, method: str, request: Message) -> Op:
        """Queue a single batch operation. Only the non-getter operations will be
        queued.

        Parameters
        ----------
        package : str
            gRPC package name
        service : str
            gRPC service name
        method : str
            gRPC method name
        request : Any
            gRPC request message

        Returns
        -------
        BatchOps.Op
            BatchOps.Op object with a queued attribute which is true if the operation
            has been queued.
        """
        op = BatchOps.Op(package, service, method, request.SerializeToString())
        if op._supported:
            network_logger.debug(
                f"Adding batch operation with package {package}, service {service} and method {method}"
            )
            self._ops.append(op)
            op.queued = True
        return op

    def clear_ops(self) -> None:
        """Clear all queued batch operations."""
        self._ops.clear()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provide a module to test the algorithms which parse job scheduler environments for
machines to run on."""

from builtins import range
import os
import socket
import tempfile
import unittest

from ansys.fluent.core.scheduler import build_parallel_options
from ansys.fluent.core.scheduler.load_machines import (
    _construct_machine_list_slurm,
    _parse_host_info,
    _parse_machine_data,
    _restrict_machines_to_core_count,
    load_machines,
)
from ansys.fluent.core.scheduler.machine_list import Machine, MachineList


class TestMachine(unittest.TestCase):
    """A basic test that checks Machine object behavior."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_initialize_host(self):
        """Test that a Machine initializes as expected."""
        machine = Machine("machine", 20)
        self.assertEqual(machine.host_name, "machine")
        self.assertEqual(machine.number_of_cores, 20)
        self.assertEqual(machine.queue_name, None)
        self.assertEqual(machine.core_list, None)

    def test_modify_host(self):
        """Test that a Machine can be modified."""
        machine = Machine("machine", 20, "allq", "0:0")
        machine.number_of_cores = 12
        self.assertEqual(machine.number_of_cores, 12)
        machine.number_of_cores = machine.number_of_cores + 2
        machine.number_of_cores += 1
        self.assertEqual(machine.number_of_cores, 15)
        machine.number_of_cores -= 3
        self.assertEqual(machine.number_of_cores, 12)


class TestMachineList(unittest.TestCase):
    """Provide a test suite that checks that the MachineList object behaves properly."""

    def setUp(self):
        self._machineList = MachineList()

    def tearDown(self):
        self._machineList.reset()

    def test_initialize_machinelist(self):
        """Tests that a host file object initializes properly."""
        newMachineFile = MachineList()
        self.assertIsInstance(newMachineFile, MachineList)
        self.assertEqual(newMachineFile.machines, [])
        self.assertEqual(newMachineFile.num_machines, 0)

    def test_copy_machinelist(self):
        """Tests that the internal copy function works properly."""
        import copy

        newMachineList = copy.deepcopy(self._machineList)
        for m1, m2 in zip(self._machineList.machines, newMachineList.machines):
            self.assertEqual(m1.host_name, m2.host_name)
            self.assertEqual(m1.number_of_cores, m2.number_of_cores)

    def test_add_to_machinelist(self):
        """Tests that a machines can be added to a machine list."""
        self._machineList.add(Machine("machine1", 20, "allq", "0:0"))
        self._machineList.add(Machine("machine2", 20, "allq", "0:0"))
        self.assertEqual(self._machineList.num_machines, 2)

    def test_number_of_cores_and_machines(self):
        """Test that the total and max number of cores and machines is working."""
        self._machineList.add(Machine("machine1", 20, "allq", "0:0"))
        self._machineList.add(Machine("machine2", 25, "allq", "0:0"))
        self._machineList.add(Machine("machine3", 15, "allq", "0:0"))
        self.assertEqual(self._machineList.num_machines, 3)
        self.assertEqual(self._machineList.number_of_cores, 60)
        self.assertEqual(self._machineList.max_cores, 25)
        self.assertEqual(self._machineList.min_cores, 15)

    def test_sort_machine_list(self):
        """Test that the machines are sorted in order of decreasing core count."""
        self._machineList.add(Machine("machine1", 15, "allq", "0:0"))
        self._machineList.add(Machine("machine2", 10, "allq", "0:0"))
        self._machineList.add(Machine("machine3", 5, "allq", "0:0"))

        # Sort in ascending order
        self._machineList.sort_by_core_count_ascending()
        numCores = self._machineList.machines[0].number_of_cores
        for h in range(1, len(self._machineList.machines)):
            self.assertLessEqual(
                numCores, self._machineList.machines[h].number_of_cores
            )
            numCores = self._machineList.machines[h].number_of_cores

        # Sort in descending order
        self._machineList.sort_by_core_count()
        numCores = self._machineList.machines[0].number_of_cores
        for h in range(1, len(self._machineList.machines)):
            self.assertLessEqual(
                self._machineList.machines[h].number_of_cores, numCores
            )
            numCores = self._machineList.machines[h].number_of_cores

    def test_remote_empty_machines(self):
        self._machineList.add(Machine("machine1", 5))
        self._machineList.add(Machine("machine2", 0))
        self._machineList.remove_empty_machines()
        self.assertEqual(self._machineList.num_machines, 1)
        self.assertEqual(self._machineList.machines[0].host_name, "machine1")

    def test_move_local_host_to_front(self):
        import socket

        localHostName = socket.gethostname()

        self._machineList.add(Machine("M0", 2))
        self._machineList.add(Machine(localHostName, 1))
        self._machineList.add(Machine("M1", 3))
        self._machineList.move_local_host_to_front()
        self.assertEqual(self._machineList.machines[0].host_name, localHostName)
        self.assertEqual(self._machineList.machines[0].number_of_cores, 1)
        self.assertEqual(self._machineList.machines[1].host_name, "M0")
        self.assertEqual(self._machineList.machines[1].number_of_cores, 2)
        self.assertEqual(self._machineList.machines[2].host_name, "M1")
        self.assertEqual(self._machineList.machines[2].number_of_cores, 3)

    def test_deep_copy_machinelist(self):
        self._machineList.add(Machine("wathpc-2-0.local", 23))
        self._machineList.add(Machine("wathpc-2-1.local", 23))
        self._machineList.add(Machine("wathpc-2-2.local", 23))
        self._machineList.add(Machine("wathpc-2-3.local", 23))
        import copy

        machineListCopy = copy.deepcopy(self._machineList)
        for m1, m2 in zip(self._machineList.machines, machineListCopy.machines):
            self.assertEqual(m1.host_name, m2.host_name)
            self.assertEqual(m1.number_of_cores, m2.number_of_cores)


class TestLoadMachines(unittest.TestCase):
    """Provide a test suite that checks that loadMachines behaves properly."""

    def setUp(self):
        self._machineList = MachineList()

    def tearDown(self):
        self._machineList.reset()

    def test_machine_info(self):
        info = [
            {"machine-name": "M0", "core-count": 1},
            {"machine-name": "M1", "core-count": 6},
        ]
        machineList = load_machines(machine_info=info)
        self.assertEqual(machineList.number_of_cores, 7)

    def test_restrict_machines(self):
        info = [
            {"machine-name": "M0", "core-count": 1},
            {"machine-name": "M1", "core-count": 1},
        ]
        machineList = load_machines(machine_info=info)
        old_machine_list = _restrict_machines_to_core_count(
            machineList, ncores=machineList.number_of_cores
        )
        self.assertEqual(machineList, old_machine_list)

    def test_pe_hostfile(self):
        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(b"m1\r\n\r\nm2 3 None None\r\nm3 4\r\nm4 2 queueName1")
            os.environ["PE_HOSTFILE"] = fp.name
        machineList = load_machines()
        os.unlink(fp.name)
        self.assertEqual(machineList.number_of_cores, 10)
        self.assertEqual(machineList.machines[1].host_name, "m2")
        self.assertEqual(machineList.machines[2].number_of_cores, 4)
        self.assertEqual(machineList.machines[3].queue_name, "queueName1")
        del os.environ["PE_HOSTFILE"]

    def test_lsb_mcpu(self):
        os.environ["LSB_MCPU_HOSTS"] = "m1 3 m2 3"
        machineList = load_machines()
        self.assertEqual(machineList.number_of_cores, 6)
        del os.environ["LSB_MCPU_HOSTS"]

    def test_pbs_nodefile(self):
        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(b"m1\r\n\r\nm2\r\nm2\r\nm2")
            os.environ["PBS_NODEFILE"] = fp.name
        machineList = load_machines()
        os.unlink(fp.name)
        self.assertEqual(machineList[1].number_of_cores, 3)
        del os.environ["PBS_NODEFILE"]

    def test_no_environment(self):
        machineList = load_machines()
        self.assertEqual(machineList[0].host_name, socket.gethostname())
        self.assertEqual(machineList.number_of_cores, 1)

    def test_no_environment_cores(self):
        machineList = load_machines(ncores=4)
        self.assertEqual(machineList[0].host_name, socket.gethostname())
        self.assertEqual(machineList.number_of_cores, 4)

    def test_constrain_machines1(self):
        machineList = load_machines(host_info="M0:2,M1:3,M2:2", ncores=4)
        expectedValue = {"M0": 1, "M1": 2, "M2": 1}
        self.assertEqual(len(machineList.machines), 3)
        for machine in machineList.machines:
            self.assertEqual(machine.number_of_cores, expectedValue[machine.host_name])
        # Ensure that the order is preserved
        self.assertEqual(machineList.machines[0].host_name, "M0")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t4 -cnf=M0:1,M1:2,M2:1")

    def test_constrain_machines2(self):
        with tempfile.NamedTemporaryFile(delete=False) as fp:
            fp.write(b"M0:2,M1:3,M2:2")
        machineList = load_machines(host_info=fp.name, ncores=3)
        expectedValue = {"M0": 1, "M1": 1, "M2": 1}
        os.unlink(fp.name)
        self.assertEqual(len(machineList.machines), 3)
        for machine in machineList.machines:
            self.assertEqual(machine.number_of_cores, expectedValue[machine.host_name])
        # Ensure that the order is preserved
        self.assertEqual(machineList.machines[0].host_name, "M0")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t3 -cnf=M0:1,M1:1,M2:1")

    def test_overload_machines1(self):
        machineList = load_machines(host_info="M0:2,M1:1", ncores=10)
        expectedValue = {"M0": 2, "M1": 1}
        self.assertEqual(len(machineList.machines), 2)
        for machine in machineList.machines:
            self.assertEqual(machine.number_of_cores, expectedValue[machine.host_name])
        # Ensure that the order is preserved
        self.assertEqual(machineList.machines[0].host_name, "M0")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t3 -cnf=M0:2,M1:1")

    def test_overload_machines2(self):
        machineList = load_machines(host_info="M0,M0,M1", ncores=10)
        expectedValue = {"M0": 2, "M1": 1}
        self.assertEqual(len(machineList.machines), 2)
        for machine in machineList.machines:
            self.assertEqual(machine.number_of_cores, expectedValue[machine.host_name])
        # Ensure that the order is preserved
        self.assertEqual(machineList.machines[0].host_name, "M0")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t3 -cnf=M0:2,M1:1")

    def test_winhpc(self):
        os.environ["CCP_NODES"] = "3 M0 8 M1 8 M2 16"
        machineList = load_machines()
        self.assertEqual(machineList.num_machines, 3)
        self.assertEqual(machineList.number_of_cores, 32)
        self.assertEqual(machineList.machines[0].host_name, "M0")
        self.assertEqual(machineList.machines[1].host_name, "M1")
        self.assertEqual(machineList.machines[2].host_name, "M2")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t32 -cnf=M0:8,M1:8,M2:16")
        del os.environ["CCP_NODES"]

    def test_slurm_single_num(self):
        os.environ["SLURM_JOB_NODELIST"] = "M[1-2],M[3]"
        os.environ["SLURM_TASKS_PER_NODE"] = "8,10(x2)"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList[1].number_of_cores, 10)
        self.assertEqual(machineList[2].number_of_cores, 10)
        self.assertEqual(machineList[2].host_name, "M3")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_TASKS_PER_NODE"]

    def test_slurm_no_brackets(self):
        os.environ["SLURM_JOB_NODELIST"] = "M0,M1,M2"
        os.environ["SLURM_NTASKS_PER_NODE"] = "8"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 3)
        self.assertEqual(machineList.number_of_cores, 24)
        self.assertEqual(machineList.machines[0].host_name, "M0")
        self.assertEqual(machineList.machines[1].host_name, "M1")
        self.assertEqual(machineList.machines[2].host_name, "M2")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t24 -cnf=M0:8,M1:8,M2:8")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]

    def test_slurm_no_padding(self):
        os.environ["SLURM_JOB_NODELIST"] = "M[0-2]"
        os.environ["SLURM_NTASKS_PER_NODE"] = "12"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 3)
        self.assertEqual(machineList.number_of_cores, 36)
        self.assertEqual(machineList.machines[0].host_name, "M0")
        self.assertEqual(machineList.machines[1].host_name, "M1")
        self.assertEqual(machineList.machines[2].host_name, "M2")
        fluentOpts = build_parallel_options(machineList)
        self.assertEqual(fluentOpts, "-t36 -cnf=M0:12,M1:12,M2:12")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]

    def test_slurm_hosts_with_dash(self):
        os.environ["SLURM_JOB_NODELIST"] = "M-n50-[0-1],M-p50-[9-11]"
        os.environ["SLURM_NTASKS_PER_NODE"] = "12"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 5)
        self.assertEqual(machineList.number_of_cores, 60)
        self.assertEqual(machineList.machines[0].host_name, "M-n50-0")
        self.assertEqual(machineList.machines[1].host_name, "M-n50-1")
        self.assertEqual(machineList.machines[2].host_name, "M-p50-9")
        self.assertEqual(machineList.machines[3].host_name, "M-p50-10")
        self.assertEqual(machineList.machines[4].host_name, "M-p50-11")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]

    def test_slurm_with_padding(self):
        os.environ["SLURM_JOB_NODELIST"] = "MC[008-009,010,011,012-014]"
        os.environ["SLURM_TASKS_PER_NODE"] = "8,10(x2),12(x3),10"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 7)
        self.assertEqual(machineList.number_of_cores, 74)
        self.assertEqual(machineList.machines[0].host_name, "MC008")
        self.assertEqual(machineList.machines[0].number_of_cores, 8)
        self.assertEqual(machineList.machines[1].host_name, "MC009")
        self.assertEqual(machineList.machines[1].number_of_cores, 10)
        self.assertEqual(machineList.machines[2].host_name, "MC010")
        self.assertEqual(machineList.machines[2].number_of_cores, 10)
        self.assertEqual(machineList.machines[3].host_name, "MC011")
        self.assertEqual(machineList.machines[3].number_of_cores, 12)
        self.assertEqual(machineList.machines[4].host_name, "MC012")
        self.assertEqual(machineList.machines[4].number_of_cores, 12)
        self.assertEqual(machineList.machines[5].host_name, "MC013")
        self.assertEqual(machineList.machines[5].number_of_cores, 12)
        self.assertEqual(machineList.machines[6].host_name, "MC014")
        self.assertEqual(machineList.machines[6].number_of_cores, 10)
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_TASKS_PER_NODE"]

    def test_slurm_with_padding_one_hostlist(self):
        os.environ["SLURM_JOB_NODELIST"] = "MD[099-101]"
        os.environ["SLURM_NTASKS_PER_NODE"] = "12"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 3)
        self.assertEqual(machineList.number_of_cores, 36)
        self.assertEqual(machineList.machines[0].host_name, "MD099")
        self.assertEqual(machineList.machines[1].host_name, "MD100")
        self.assertEqual(machineList.machines[2].host_name, "MD101")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]

    def test_slurm_no_padding_commas(self):
        os.environ["SLURM_JOB_NODELIST"] = "M[2-3,4,5-7,8-11,12-14,15-16]"
        os.environ["SLURM_NTASKS_PER_NODE"] = "12"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 15)
        self.assertEqual(machineList.number_of_cores, 180)
        self.assertEqual(machineList.machines[0].host_name, "M2")
        self.assertEqual(machineList.machines[1].host_name, "M3")
        self.assertEqual(machineList.machines[2].host_name, "M4")
        self.assertEqual(machineList.machines[3].host_name, "M5")
        self.assertEqual(machineList.machines[4].host_name, "M6")
        self.assertEqual(machineList.machines[5].host_name, "M7")
        self.assertEqual(machineList.machines[6].host_name, "M8")
        self.assertEqual(machineList.machines[7].host_name, "M9")
        self.assertEqual(machineList.machines[8].host_name, "M10")
        self.assertEqual(machineList.machines[9].host_name, "M11")
        self.assertEqual(machineList.machines[10].host_name, "M12")
        self.assertEqual(machineList.machines[11].host_name, "M13")
        self.assertEqual(machineList.machines[12].host_name, "M14")
        self.assertEqual(machineList.machines[13].host_name, "M15")
        self.assertEqual(machineList.machines[14].host_name, "M16")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]

    def test_slurm_very_complex(self):
        os.environ["SLURM_JOB_NODELIST"] = (
            "M[2-3,4,5-7,8-11,12-14,15-16],MB,MC[008-009,010-011,012-014],MD[099-101]"
        )
        os.environ["SLURM_NTASKS_PER_NODE"] = "24"
        hostList = os.environ.get("SLURM_JOB_NODELIST")
        machineList = _construct_machine_list_slurm(hostList)
        self.assertEqual(machineList.num_machines, 26)
        self.assertEqual(machineList.number_of_cores, 624)
        self.assertEqual(machineList.machines[0].host_name, "M2")
        self.assertEqual(machineList.machines[0].number_of_cores, 24)
        self.assertEqual(machineList.machines[7].host_name, "M9")
        self.assertEqual(machineList.machines[14].host_name, "M16")
        self.assertEqual(machineList.machines[15].host_name, "MB")
        self.assertEqual(machineList.machines[16].host_name, "MC008")
        self.assertEqual(machineList.machines[19].host_name, "MC011")
        self.assertEqual(machineList.machines[22].host_name, "MC014")
        self.assertEqual(machineList.machines[24].host_name, "MD100")
        del os.environ["SLURM_JOB_NODELIST"]
        del os.environ["SLURM_NTASKS_PER_NODE"]


class TestMachineListCmdLine(unittest.TestCase):
    """Provide a test suite that checks the machine list parser."""

    def setUp(self):
        self._expectedValues = {"M0": 2, "M1": 4}

    def tearDown(self):
        pass

    def test_parse_machine_data(self):
        machineDataList = [["M0:2", "M1:2", "M1:2"], ["M0", "M0", "M1", "M1:3"]]

        for machineData in machineDataList:
            machineList = _parse_machine_data(machineData)
            for machine in machineList.machines:
                self.assertEqual(
                    machine.number_of_cores, self._expectedValues[machine.host_name]
                )

    def test_cmd_string(self):
        hostLists = ["M0:2,M1:2,M1:2", "M0,M0,M1,M1:3"]

        for hostList in hostLists:
            machineList = _parse_host_info(hostList)
            for machine in machineList.machines:
                self.assertEqual(
                    machine.number_of_cores, self._expectedValues[machine.host_name]
                )

    def test_no_machine_name(self):
        hostList = "M0:2,M1:2,"
        with self.assertRaises(RuntimeError) as cm:
            _parse_host_info(hostList)
        self.assertEqual(str(cm.exception), "Problem with machine list format.")

    def test_host_file(self):
        import os.path

        hostfile = "hosts.txt"
        # This unit test only runs if the file exists
        if os.path.isfile(hostfile):
            machineList = _parse_host_info(hostfile)
            for machine in machineList.machines:
                self.assertEqual(
                    machine.number_of_cores, self._expectedValues[machine.host_name]
                )


suite1 = unittest.TestLoader().loadTestsFromTestCase(TestMachine)
suite2 = unittest.TestLoader().loadTestsFromTestCase(TestMachineList)
suite3 = unittest.TestLoader().loadTestsFromTestCase(TestLoadMachines)
suite4 = unittest.TestLoader().loadTestsFromTestCase(TestMachineListCmdLine)
alltests = unittest.TestSuite([suite1, suite2, suite3, suite4])
unittest.TextTestRunner(verbosity=2).run(alltests)

if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for datamodel streaming."""

import logging

from google.protobuf.json_format import MessageToDict

from ansys.api.fluent.v0 import datamodel_se_pb2
import ansys.fluent.core as pyfluent
from ansys.fluent.core.streaming_services.streaming import StreamingService

network_logger: logging.Logger = logging.getLogger("pyfluent.networking")


class DatamodelStream(StreamingService):
    """Encapsulates a datamodel streaming service."""

    def __init__(self, service):
        """Initialize DatamodelStream."""
        super().__init__(
            stream_begin_method="BeginStreaming",
            target=DatamodelStream._process_streaming,
            streaming_service=service,
        )

    def _process_streaming(
        self,
        id,
        stream_begin_method,
        started_evt,
        rules,
        no_commands_diff_state,
        *args,
        **kwargs,
    ):
        """Processes datamodel events."""
        data_model_request = datamodel_se_pb2.DataModelRequest(*args, **kwargs)
        data_model_request.rules = rules
        data_model_request.returnstatechanges = pyfluent.DATAMODEL_RETURN_STATE_CHANGES
        if no_commands_diff_state:
            data_model_request.diffstate = datamodel_se_pb2.DIFFSTATE_NOCOMMANDS
        responses = self._streaming_service.begin_streaming(
            data_model_request,
            started_evt,
            id=id,
            stream_begin_method=stream_begin_method,
        )
        while True:
            try:
                response: datamodel_se_pb2.DataModelResponse = next(responses)
                network_logger.debug(
                    f"GRPC_TRACE: RPC = /grpcRemoting.DataModel/BeginStreaming, response = {MessageToDict(response)}"
                )
                with self._lock:
                    self._streaming = True
                    for _, cb_list in self._service_callbacks.items():
                        state = response.state if hasattr(response, "state") else None
                        deleted_paths = getattr(response, "deletedpaths", None)
                        cb_list[0](state=state, deleted_paths=deleted_paths)
            except StopIteration:
                break



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
End-to-end Fluent Solver Workflow using Watertight Meshing
-----------------------------------------------------------------------------
This test covers the setup and solution of a 3D
turbulent fluid flow and heat transfer problem in a mixing elbow. The mixing
elbow configuration is encountered in piping systems in power plants and
process industries. It is often important to predict the flow field and
temperature field in the area of the mixing region to properly design
the junction.

This test queries the following using PyTest:

- Meshing workflow tasks state before and after the task execution
- Report definitions check after solution
"""

from functools import partial

import pytest
from util.meshing_workflow import (
    assign_task_arguments,
    execute_task_with_pre_and_postcondition_checks,
)
from util.solver import check_report_definition_result

from ansys.fluent.core.utils.fluent_version import FluentVersion


@pytest.mark.nightly
@pytest.mark.codegen_required
def test_mixing_elbow(
    watertight_workflow_session_wo_exit, mixing_elbow_geometry_filename
):
    meshing_session = watertight_workflow_session_wo_exit
    workflow = meshing_session.workflow

    assign_task_args = partial(
        assign_task_arguments, workflow=workflow, check_state=True
    )

    execute_task_with_pre_and_postconditions = partial(
        execute_task_with_pre_and_postcondition_checks, workflow=workflow
    )

    ###############################################################################
    # Import the CAD geometry
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Import Geometry",
        FileName=mixing_elbow_geometry_filename,
        LengthUnit="in",
    )

    execute_task_with_pre_and_postconditions(task_name="Import Geometry")

    ###############################################################################
    # Add local sizing
    # Query the task state before and after task execution
    workflow.TaskObject["Add Local Sizing"].AddChildToTask()

    execute_task_with_pre_and_postconditions(task_name="Add Local Sizing")

    ###############################################################################
    # Generate the surface mesh
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Generate the Surface Mesh",
        CFDSurfaceMeshControls={"MaxSize": 0.3},
    )

    execute_task_with_pre_and_postconditions(task_name="Generate the Surface Mesh")

    ###############################################################################
    # Describe the geometry
    # Query the task state before and after task execution
    describe_geo = workflow.TaskObject["Describe Geometry"]
    describe_geo.UpdateChildTasks(SetupTypeChanged=False)
    assign_task_args(
        task_name="Describe Geometry",
        SetupType="The geometry consists of only fluid regions with no voids",
    )
    describe_geo.UpdateChildTasks(SetupTypeChanged=True)

    execute_task_with_pre_and_postconditions(task_name="Describe Geometry")

    ###############################################################################
    # Update Boundaries Task
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Update Boundaries",
        BoundaryLabelList=["wall-inlet"],
        BoundaryLabelTypeList=["wall"],
        OldBoundaryLabelList=["wall-inlet"],
        OldBoundaryLabelTypeList=["velocity-inlet"],
    )

    execute_task_with_pre_and_postconditions(task_name="Update Boundaries")

    ###############################################################################
    # Update your regions
    # Query the task state before and after task execution

    execute_task_with_pre_and_postconditions(task_name="Update Regions")

    ###############################################################################
    # Add Boundary Layers
    # Query the task state before and after task execution
    add_boundary_layers = workflow.TaskObject["Add Boundary Layers"]
    add_boundary_layers.AddChildToTask()
    add_boundary_layers.InsertCompoundChildTask()
    assign_task_args(
        task_name="smooth-transition_1", BLControlName="smooth-transition_1"
    )
    add_boundary_layers.Arguments = {}

    execute_task_with_pre_and_postconditions(task_name="Add Boundary Layers")

    ###############################################################################
    # Generate the volume mesh
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Generate the Volume Mesh",
        VolumeFill="poly-hexcore",
        VolumeFillControls={"HexMaxCellLength": 0.3},
    )

    execute_task_with_pre_and_postconditions(task_name="Generate the Volume Mesh")

    ###############################################################################
    # Check the mesh in Meshing mode
    # TODO: Remove the if condition after a stable version of 23.1 is available and update the commands as required.
    if meshing_session.get_fluent_version() < FluentVersion.v231:
        meshing_session.tui.mesh.check_mesh()

    ###############################################################################
    # Switch to Solution mode
    solver_session = meshing_session.switch_to_solver()

    ###############################################################################
    # Check the mesh in Solver mode
    solver_session.tui.mesh.check()

    ###############################################################################
    # Set the working units for the mesh
    define = solver_session.tui.define
    define.units("length", "in")

    ###############################################################################
    # Enable heat transfer by activating the energy equation.
    define.models.energy("yes", ", ", ", ", ", ", ", ")

    ###############################################################################
    # Create a new material called water-liquid.
    define.materials.copy("fluid", "water-liquid")

    ###############################################################################
    # Set up the cell zone conditions for the fluid zone (elbow-fluid). Select
    # water-liquid from the Material list.
    # TODO: Remove the if condition after a stable version of 23.1 is available and update the commands as required.
    if solver_session.get_fluent_version() < FluentVersion.v231:
        define.boundary_conditions.fluid(
            "elbow-fluid",
            "yes",
            "water-liquid",
            "no",
            "no",
            "no",
            "no",
            "0",
            "no",
            "0",
            "no",
            "0",
            "no",
            "0",
            "no",
            "0",
            "no",
            "1",
            "no",
            "no",
            "no",
            "no",
            "no",
        )

        ###############################################################################
        # Set up the boundary conditions
        velocity_inlet = define.boundary_conditions.set.velocity_inlet
        velocity_inlet("cold-inlet", [], "vmag", "no", 0.4, "quit")
        velocity_inlet("cold-inlet", [], "ke-spec", "no", "no", "no", "yes", "quit")
        velocity_inlet("cold-inlet", [], "turb-intensity", 5, "quit")
        velocity_inlet("cold-inlet", [], "turb-hydraulic-diam", 4, "quit")
        velocity_inlet("cold-inlet", [], "temperature", "no", 293.15, "quit")
        velocity_inlet("hot-inlet", [], "vmag", "no", 1.2, "quit")
        velocity_inlet("hot-inlet", [], "ke-spec", "no", "no", "no", "yes", "quit")
        velocity_inlet("hot-inlet", [], "turb-intensity", 5, "quit")
        velocity_inlet("hot-inlet", [], "turb-hydraulic-diam", 1, "quit")
        velocity_inlet("hot-inlet", [], "temperature", "no", 313.15, "quit")

        pressure_outlet = define.boundary_conditions.set.pressure_outlet
        pressure_outlet("outlet", [], "turb-intensity", 5, "quit")
        pressure_outlet("outlet", [], "turb-viscosity-ratio", 4, "quit")

        ###############################################################################
        # Enable the plotting of residuals during the calculation.
        solver_session.tui.solve.monitors.residual.plot("yes")

        ###############################################################################
        # Initialize the flow field using the Hybrid Initialization
        solver_session.tui.solve.initialize.hyb_initialization()

        ###############################################################################
        # Solve for 250 Iterations.
        solver_session.tui.solve.iterate(250)

        ###############################################################################
        # Assert the returned mass flow rate report definition value
        flux = solver_session.solution.report_definitions.flux
        flux["mass_flow_rate"] = {}
        flux["mass_flow_rate"].zone_names = [
            "cold-inlet",
            "hot-inlet",
            "outlet",
        ]

        check_report_definition = partial(
            check_report_definition_result,
            report_definitions=solver_session.solution.report_definitions,
        )

        check_report_definition(
            report_definition_name="mass_flow_rate",
            expected_result=pytest.approx(-2.985690364942784e-06, abs=1e-3),
        )

        ###############################################################################
        # Assert the returned temperature report definition value on the outlet surface
        solver_session.solution.report_definitions.surface["temperature_outlet"] = {}
        temp_outlet = solver_session.solution.report_definitions.surface[
            "temperature_outlet"
        ]
        temp_outlet.report_type = "surface-massavg"
        temp_outlet.field = "temperature"
        temp_outlet.surface_names = ["outlet"]

        check_report_definition(
            report_definition_name="temperature_outlet",
            expected_result=pytest.approx(296.229, rel=1e-3),
        )

        ###############################################################################
    solver_session.exit()



#
# This is an auto-generated file.  DO NOT EDIT!
#
# pylint: disable=line-too-long

from ansys.fluent.core.services.datamodel_se import (
    PyMenu,
    PyParameter,
    PyTextual,
    PyNumerical,
    PyDictionary,
    PyNamedObjectContainer,
    PyCommand,
    PyQuery,
    PyCommandArguments,
    PyTextualCommandArgumentsSubItem,
    PyNumericalCommandArgumentsSubItem,
    PyDictionaryCommandArgumentsSubItem,
    PyParameterCommandArgumentsSubItem,
    PySingletonCommandArgumentsSubItem
)


class Root(PyMenu):
    """
    Singleton Root.
    """
    def __init__(self, service, rules, path):
        self.TaskObject = self.__class__.TaskObject(service, rules, path + [("TaskObject", "")])
        self.Workflow = self.__class__.Workflow(service, rules, path + [("Workflow", "")])
        self.CreateCompositeTask = self.__class__.CreateCompositeTask(service, rules, "CreateCompositeTask", path)
        self.CreateNewWorkflow = self.__class__.CreateNewWorkflow(service, rules, "CreateNewWorkflow", path)
        self.DeleteTasks = self.__class__.DeleteTasks(service, rules, "DeleteTasks", path)
        self.InitializeWorkflow = self.__class__.InitializeWorkflow(service, rules, "InitializeWorkflow", path)
        self.InsertNewTask = self.__class__.InsertNewTask(service, rules, "InsertNewTask", path)
        self.LoadState = self.__class__.LoadState(service, rules, "LoadState", path)
        self.LoadWorkflow = self.__class__.LoadWorkflow(service, rules, "LoadWorkflow", path)
        self.ResetWorkflow = self.__class__.ResetWorkflow(service, rules, "ResetWorkflow", path)
        self.SaveWorkflow = self.__class__.SaveWorkflow(service, rules, "SaveWorkflow", path)
        super().__init__(service, rules, path)

    class TaskObject(PyNamedObjectContainer):
        """
        .
        """
        class _TaskObject(PyMenu):
            """
            Singleton _TaskObject.
            """
            def __init__(self, service, rules, path):
                self.Arguments = self.__class__.Arguments(service, rules, path + [("Arguments", "")])
                self.CheckPoint = self.__class__.CheckPoint(service, rules, path + [("CheckPoint", "")])
                self.CommandName = self.__class__.CommandName(service, rules, path + [("CommandName", "")])
                self.Errors = self.__class__.Errors(service, rules, path + [("Errors", "")])
                self.InactiveTaskList = self.__class__.InactiveTaskList(service, rules, path + [("InactiveTaskList", "")])
                self.ObjectPath = self.__class__.ObjectPath(service, rules, path + [("ObjectPath", "")])
                self.State = self.__class__.State(service, rules, path + [("State", "")])
                self.TaskList = self.__class__.TaskList(service, rules, path + [("TaskList", "")])
                self.TaskType = self.__class__.TaskType(service, rules, path + [("TaskType", "")])
                self.Warnings = self.__class__.Warnings(service, rules, path + [("Warnings", "")])
                self._name_ = self.__class__._name_(service, rules, path + [("_name_", "")])
                self.AddChildAndUpdate = self.__class__.AddChildAndUpdate(service, rules, "AddChildAndUpdate", path)
                self.AddChildToTask = self.__class__.AddChildToTask(service, rules, "AddChildToTask", path)
                self.DisplayModel = self.__class__.DisplayModel(service, rules, "DisplayModel", path)
                self.Execute = self.__class__.Execute(service, rules, "Execute", path)
                self.ExecuteUpstreamNonExecutedAndThisTask = self.__class__.ExecuteUpstreamNonExecutedAndThisTask(service, rules, "ExecuteUpstreamNonExecutedAndThisTask", path)
                self.ForceUptoDate = self.__class__.ForceUptoDate(service, rules, "ForceUptoDate", path)
                self.GetNextPossibleTasks = self.__class__.GetNextPossibleTasks(service, rules, "GetNextPossibleTasks", path)
                self.InsertCompositeChildTask = self.__class__.InsertCompositeChildTask(service, rules, "InsertCompositeChildTask", path)
                self.InsertCompoundChildTask = self.__class__.InsertCompoundChildTask(service, rules, "InsertCompoundChildTask", path)
                self.InsertNextTask = self.__class__.InsertNextTask(service, rules, "InsertNextTask", path)
                self.Rename = self.__class__.Rename(service, rules, "Rename", path)
                self.Revert = self.__class__.Revert(service, rules, "Revert", path)
                self.SetAsCurrent = self.__class__.SetAsCurrent(service, rules, "SetAsCurrent", path)
                self.UpdateChildTasks = self.__class__.UpdateChildTasks(service, rules, "UpdateChildTasks", path)
                super().__init__(service, rules, path)

            class Arguments(PyDictionary):
                """
                Parameter Arguments of value type dict[str, Any].
                """
                pass

            class CheckPoint(PyTextual):
                """
                Parameter CheckPoint of value type str.
                """
                pass

            class CommandName(PyTextual):
                """
                Parameter CommandName of value type str.
                """
                pass

            class Errors(PyTextual):
                """
                Parameter Errors of value type list[str].
                """
                pass

            class InactiveTaskList(PyTextual):
                """
                Parameter InactiveTaskList of value type list[str].
                """
                pass

            class ObjectPath(PyTextual):
                """
                Parameter ObjectPath of value type str.
                """
                pass

            class State(PyTextual):
                """
                Parameter State of value type str.
                """
                pass

            class TaskList(PyTextual):
                """
                Parameter TaskList of value type list[str].
                """
                pass

            class TaskType(PyTextual):
                """
                Parameter TaskType of value type str.
                """
                pass

            class Warnings(PyTextual):
                """
                Parameter Warnings of value type list[str].
                """
                pass

            class _name_(PyTextual):
                """
                Parameter _name_ of value type str.
                """
                pass

            class AddChildAndUpdate(PyCommand):
                """
                Command AddChildAndUpdate.

                Parameters
                ----------
                DeferUpdate : bool
                Force : bool

                Returns
                -------
                bool
                """
                class _AddChildAndUpdateCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.DeferUpdate = self._DeferUpdate(self, "DeferUpdate", service, rules, path)
                        self.Force = self._Force(self, "Force", service, rules, path)

                    class _DeferUpdate(PyParameterCommandArgumentsSubItem):
                        """
                        Argument DeferUpdate.
                        """

                    class _Force(PyParameterCommandArgumentsSubItem):
                        """
                        Argument Force.
                        """

                def create_instance(self) -> _AddChildAndUpdateCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._AddChildAndUpdateCommandArguments(*args)

            class AddChildToTask(PyCommand):
                """
                Command AddChildToTask.


                Returns
                -------
                bool
                """
                class _AddChildToTaskCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _AddChildToTaskCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._AddChildToTaskCommandArguments(*args)

            class DisplayModel(PyCommand):
                """
                Command DisplayModel.


                Returns
                -------
                bool
                """
                class _DisplayModelCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _DisplayModelCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._DisplayModelCommandArguments(*args)

            class Execute(PyCommand):
                """
                Command Execute.

                Parameters
                ----------
                Force : bool

                Returns
                -------
                bool
                """
                class _ExecuteCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.Force = self._Force(self, "Force", service, rules, path)

                    class _Force(PyParameterCommandArgumentsSubItem):
                        """
                        Argument Force.
                        """

                def create_instance(self) -> _ExecuteCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._ExecuteCommandArguments(*args)

            class ExecuteUpstreamNonExecutedAndThisTask(PyCommand):
                """
                Command ExecuteUpstreamNonExecutedAndThisTask.


                Returns
                -------
                bool
                """
                class _ExecuteUpstreamNonExecutedAndThisTaskCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _ExecuteUpstreamNonExecutedAndThisTaskCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._ExecuteUpstreamNonExecutedAndThisTaskCommandArguments(*args)

            class ForceUptoDate(PyCommand):
                """
                Command ForceUptoDate.

                Parameters
                ----------
                Force : bool

                Returns
                -------
                bool
                """
                class _ForceUptoDateCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.Force = self._Force(self, "Force", service, rules, path)

                    class _Force(PyParameterCommandArgumentsSubItem):
                        """
                        Argument Force.
                        """

                def create_instance(self) -> _ForceUptoDateCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._ForceUptoDateCommandArguments(*args)

            class GetNextPossibleTasks(PyCommand):
                """
                Command GetNextPossibleTasks.


                Returns
                -------
                bool
                """
                class _GetNextPossibleTasksCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _GetNextPossibleTasksCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._GetNextPossibleTasksCommandArguments(*args)

            class InsertCompositeChildTask(PyCommand):
                """
                Command InsertCompositeChildTask.

                Parameters
                ----------
                CommandName : str

                Returns
                -------
                bool
                """
                class _InsertCompositeChildTaskCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.CommandName = self._CommandName(self, "CommandName", service, rules, path)

                    class _CommandName(PyTextualCommandArgumentsSubItem):
                        """
                        Argument CommandName.
                        """

                def create_instance(self) -> _InsertCompositeChildTaskCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._InsertCompositeChildTaskCommandArguments(*args)

            class InsertCompoundChildTask(PyCommand):
                """
                Command InsertCompoundChildTask.


                Returns
                -------
                bool
                """
                class _InsertCompoundChildTaskCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _InsertCompoundChildTaskCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._InsertCompoundChildTaskCommandArguments(*args)

            class InsertNextTask(PyCommand):
                """
                Command InsertNextTask.

                Parameters
                ----------
                CommandName : str
                Select : bool

                Returns
                -------
                bool
                """
                class _InsertNextTaskCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.CommandName = self._CommandName(self, "CommandName", service, rules, path)
                        self.Select = self._Select(self, "Select", service, rules, path)

                    class _CommandName(PyTextualCommandArgumentsSubItem):
                        """
                        Argument CommandName.
                        """

                    class _Select(PyParameterCommandArgumentsSubItem):
                        """
                        Argument Select.
                        """

                def create_instance(self) -> _InsertNextTaskCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._InsertNextTaskCommandArguments(*args)

            class Rename(PyCommand):
                """
                Command Rename.

                Parameters
                ----------
                NewName : str

                Returns
                -------
                bool
                """
                class _RenameCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.NewName = self._NewName(self, "NewName", service, rules, path)

                    class _NewName(PyTextualCommandArgumentsSubItem):
                        """
                        Argument NewName.
                        """

                def create_instance(self) -> _RenameCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._RenameCommandArguments(*args)

            class Revert(PyCommand):
                """
                Command Revert.


                Returns
                -------
                bool
                """
                class _RevertCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _RevertCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._RevertCommandArguments(*args)

            class SetAsCurrent(PyCommand):
                """
                Command SetAsCurrent.


                Returns
                -------
                bool
                """
                class _SetAsCurrentCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)

                def create_instance(self) -> _SetAsCurrentCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._SetAsCurrentCommandArguments(*args)

            class UpdateChildTasks(PyCommand):
                """
                Command UpdateChildTasks.

                Parameters
                ----------
                SetupTypeChanged : bool
                Arguments : dict[str, Any]

                Returns
                -------
                bool
                """
                class _UpdateChildTasksCommandArguments(PyCommandArguments):
                    def __init__(self, service, rules, command, path, id):
                        super().__init__(service, rules, command, path, id)
                        self.SetupTypeChanged = self._SetupTypeChanged(self, "SetupTypeChanged", service, rules, path)
                        self.Arguments = self._Arguments(self, "Arguments", service, rules, path)

                    class _SetupTypeChanged(PyParameterCommandArgumentsSubItem):
                        """
                        Argument SetupTypeChanged.
                        """

                    class _Arguments(PyDictionaryCommandArgumentsSubItem):
                        """
                        Argument Arguments.
                        """

                def create_instance(self) -> _UpdateChildTasksCommandArguments:
                    args = self._get_create_instance_args()
                    if args is not None:
                        return self._UpdateChildTasksCommandArguments(*args)

        def __getitem__(self, key: str) -> _TaskObject:
            return super().__getitem__(key)

    class Workflow(PyMenu):
        """
        Singleton Workflow.
        """
        def __init__(self, service, rules, path):
            self.CurrentTask = self.__class__.CurrentTask(service, rules, path + [("CurrentTask", "")])
            self.InactiveTaskList = self.__class__.InactiveTaskList(service, rules, path + [("InactiveTaskList", "")])
            self.TaskList = self.__class__.TaskList(service, rules, path + [("TaskList", "")])
            self.WorkflowType = self.__class__.WorkflowType(service, rules, path + [("WorkflowType", "")])
            super().__init__(service, rules, path)

        class CurrentTask(PyTextual):
            """
            Parameter CurrentTask of value type str.
            """
            pass

        class InactiveTaskList(PyTextual):
            """
            Parameter InactiveTaskList of value type list[str].
            """
            pass

        class TaskList(PyTextual):
            """
            Parameter TaskList of value type list[str].
            """
            pass

        class WorkflowType(PyTextual):
            """
            Parameter WorkflowType of value type str.
            """
            pass

    class CreateCompositeTask(PyCommand):
        """
        Command CreateCompositeTask.

        Parameters
        ----------
        ListOfTasks : list[str]

        Returns
        -------
        bool
        """
        class _CreateCompositeTaskCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.ListOfTasks = self._ListOfTasks(self, "ListOfTasks", service, rules, path)

            class _ListOfTasks(PyTextualCommandArgumentsSubItem):
                """
                Argument ListOfTasks.
                """

        def create_instance(self) -> _CreateCompositeTaskCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._CreateCompositeTaskCommandArguments(*args)

    class CreateNewWorkflow(PyCommand):
        """
        Command CreateNewWorkflow.


        Returns
        -------
        bool
        """
        class _CreateNewWorkflowCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)

        def create_instance(self) -> _CreateNewWorkflowCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._CreateNewWorkflowCommandArguments(*args)

    class DeleteTasks(PyCommand):
        """
        Command DeleteTasks.

        Parameters
        ----------
        ListOfTasks : list[str]

        Returns
        -------
        bool
        """
        class _DeleteTasksCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.ListOfTasks = self._ListOfTasks(self, "ListOfTasks", service, rules, path)

            class _ListOfTasks(PyTextualCommandArgumentsSubItem):
                """
                Argument ListOfTasks.
                """

        def create_instance(self) -> _DeleteTasksCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._DeleteTasksCommandArguments(*args)

    class InitializeWorkflow(PyCommand):
        """
        Command InitializeWorkflow.

        Parameters
        ----------
        WorkflowType : str

        Returns
        -------
        bool
        """
        class _InitializeWorkflowCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.WorkflowType = self._WorkflowType(self, "WorkflowType", service, rules, path)

            class _WorkflowType(PyTextualCommandArgumentsSubItem):
                """
                Argument WorkflowType.
                """

        def create_instance(self) -> _InitializeWorkflowCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._InitializeWorkflowCommandArguments(*args)

    class InsertNewTask(PyCommand):
        """
        Command InsertNewTask.

        Parameters
        ----------
        CommandName : str

        Returns
        -------
        bool
        """
        class _InsertNewTaskCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.CommandName = self._CommandName(self, "CommandName", service, rules, path)

            class _CommandName(PyTextualCommandArgumentsSubItem):
                """
                Argument CommandName.
                """

        def create_instance(self) -> _InsertNewTaskCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._InsertNewTaskCommandArguments(*args)

    class LoadState(PyCommand):
        """
        Command LoadState.

        Parameters
        ----------
        ListOfRoots : list[str]

        Returns
        -------
        bool
        """
        class _LoadStateCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.ListOfRoots = self._ListOfRoots(self, "ListOfRoots", service, rules, path)

            class _ListOfRoots(PyTextualCommandArgumentsSubItem):
                """
                Argument ListOfRoots.
                """

        def create_instance(self) -> _LoadStateCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._LoadStateCommandArguments(*args)

    class LoadWorkflow(PyCommand):
        """
        Command LoadWorkflow.

        Parameters
        ----------
        FilePath : str

        Returns
        -------
        bool
        """
        class _LoadWorkflowCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.FilePath = self._FilePath(self, "FilePath", service, rules, path)

            class _FilePath(PyTextualCommandArgumentsSubItem):
                """
                Argument FilePath.
                """

        def create_instance(self) -> _LoadWorkflowCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._LoadWorkflowCommandArguments(*args)

    class ResetWorkflow(PyCommand):
        """
        Command ResetWorkflow.


        Returns
        -------
        bool
        """
        class _ResetWorkflowCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)

        def create_instance(self) -> _ResetWorkflowCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._ResetWorkflowCommandArguments(*args)

    class SaveWorkflow(PyCommand):
        """
        Command SaveWorkflow.

        Parameters
        ----------
        FilePath : str

        Returns
        -------
        bool
        """
        class _SaveWorkflowCommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.FilePath = self._FilePath(self, "FilePath", service, rules, path)

            class _FilePath(PyTextualCommandArgumentsSubItem):
                """
                Argument FilePath.
                """

        def create_instance(self) -> _SaveWorkflowCommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._SaveWorkflowCommandArguments(*args)

